from encoders.utils import *
from encoders.helpers import *
from extractors.strategy import *

from extractors.solvers.msat import *
from checkers import *

import common
import random


class LazyStrategyExtractor(object):

    def __init__(self, environment):
        self.environment = environment
        self.solver = MathSAT5Solver(environment)
        return


    def getSimplex(self, yvars):
        points = []

        self.solver.push()
        for i in range(0, len(yvars)+1):
            row = []
            neg = []

            res = self.solver.solve()
            assert res
            for y in yvars:
                v = self.solver.getModelValue(y)
                row.append(v)
                neg.append(Not(Equals(y,Constant(v))))

            points.append(row)
            self.solver.addAssertion(Or(neg))

        self.solver.pop()
        return points


    def getLinearStrategy(self, simplex, clausifier, psi):
        yvars = clausifier.yvars
        xvars = clausifier.xvars
        st = self.environment.symbolTable

        s = LinearStrategy(yvars)

        tmpSolver = MathSAT5Solver(self.environment, debugFile="/tmp/dbg.smt2")

        print("    Declaring Variables...")
        coeff = []
        for _ in range(0, len(xvars)):
            row = []
            for _ in range(0, len(yvars) + 1):
                nv = st.createNewFreshSymbol(Types.REAL, "_C_%d")
                row.append(nv)
                tmpSolver.declareVariable(nv)
            coeff.append(row)

        print("    Encoding problem as SMT formula...")
        formula = self.linearEncoding(coeff, simplex, clausifier, psi)


        print("    Asserting SMT formula...")
        formula = formula.simplify(True)
        tmpSolver.addAssertion(formula)

        print("    Solving SMT formula...")
        if not tmpSolver.solve():
            return None

        print("    Extracting model...")
        for i in range(0, len(xvars)):
            row = coeff[i]
            x = xvars[i]

            vals = []
            for c in row:
                vals.append(tmpSolver.getModelValue(c))
            s.addPolicy(x, vals)

        tmpSolver.destroy()
        return s


    def linearEncoding(self, coeff, simplex, clausifier, psi):
        yvars = clausifier.yvars
        xvars = clausifier.xvars
        ycount = len(yvars)

        print("      Preparing formula definitions...")
        formula = psi

        print("      Building problem formula...")
        for i in range(0, len(xvars)):
            row = coeff[i]
            x = xvars[i]

            f = row[-1]
            for j in range(0, len(row) - 1):
                c = row[j]
                f = Plus(Times(c, yvars[j]), f)

            formula = Definition(x.name, f, formula)

        formula = formula.simplify()

        print("      Creating simplex encoding...")
        reslist = []
        for point in simplex:
            start = formula
            for i in range(0, len(point)):
                start = Definition(yvars[i].name, Constant(point[i]), start)
            reslist.append(start)

        return And(reslist)


    def getPsiWithStartegy(self, strategy, clausifier):
        xvars = clausifier.xvars
        formula = clausifier.getPsi()

        for x in xvars:
            formula = Definition(x.name, strategy.getPolicyAsFormula(x), formula)

        return (formula.simplify(True))


    def extractStrategy(self, clausifier, gamma=None, psi=None):
        # assert (gamma and psi) or clausifier.problem.disjunctivity == DisjunctivityLevel.NONE
        yvars = clausifier.yvars

        print("Starting strategy extraction...")

        s = PiecewiseStrategy(yvars)

        print("Getting initial uncontrollable region...")
        eta = clausifier.getGamma()
        if gamma is not None:
            eta = gamma

        if psi is None:
            psi = clausifier.getPsi()

        print("Declaring y variables...")
        for y in yvars:
            self.solver.declareVariable(y)

        print("Asserting initial uncontrollable region...")
        self.solver.addAssertion(eta)

        while True:
            print("\n  Checking if the remaining region is feasable...")
            if not self.solver.solve():
                print("  Region is infeasable, therefore I covered all the space!")
                print("  Done!")
                self.solver.destroy()
                return s

            print("  Region is feasible, extracting a simplex...")
            simplex = self.getSimplex(yvars)

            print("  Generating a linear startegy for the simplex...")
            p = self.getLinearStrategy(simplex, clausifier, psi)
            assert p is not None

            print("  Computing expanded region for the found strategy...")
            comb = self.getPsiWithStartegy(p, clausifier)
            comb = comb.simplify(True)
            beta = And([eta, comb])

            print("  Adding clause to global policy...")
            s.addPiece(beta, p)

            print("  Asserting out-of-region clauses...")
            self.solver.addAssertion(Not(comb))
            eta = And([eta, Not(comb)])

        return None
    pass
