import re
from mathsat import *

from formula import *
from encoders.printers import *
from extractors.solvers.converter import *

import common


class MathSAT5Solver(object):

    def __init__(self, environment, debugFile=None):
        self.declarations = {}

        self.environment = environment

        self.config = msat_create_config()
        check = msat_set_option(self.config, "model_generation", "true")
        assert check == 0

        if debugFile is not None:
            msat_set_option(self.config, "debug.api_call_trace", "1")
            msat_set_option(self.config, "debug.api_call_trace_filename", debugFile)

        #msat_set_option(self.config, "dpll.branching_initial_phase", "1")
        self.env = msat_create_env(self.config)

        self.converter = MSatConverter(self.env)
        self.realType = msat_get_rational_type(self.env)
        self.boolType = msat_get_bool_type(self.env)
        return


    def resetAssertions(self):
        msat_reset_env(self.env)
        return


    def declareVariable(self, var):
        assert var.nodeType == FNode.SYMBOL
        if var.name not in self.declarations:
            tp = None
            if var.typename == Types.BOOL:
                tp = self.boolType
            elif var.typename == Types.REAL:
                tp = self.realType

            assert tp is not None
            decl = msat_declare_function(self.env, var.name, tp)
            self.declarations[var.name] = decl

    def addAssertion(self, formula):
        # buf = common.getIOBuffer()
        # p = SmtPrinter(buf, strictSyntax=True, useBytes=True, smt2version=True)
        # p.printer(formula)
        # strformula = buf.getvalue()
        # buf.close()
        # strformula = common.convertString(strformula)
        # term = msat_from_string(self.env, strformula)

        # import time
        # time.sleep(5)

        term = self.converter.convert(formula)
        assert self.converter.safeState()

        if MSAT_ERROR_TERM(term):
            print("There was an error in formula assertion. The reported message is:")
            print("'" + msat_last_error_message(self.env) + "'")
            exit(1)

        res = msat_assert_formula(self.env, term)
        assert res == 0
        return

    def solve(self):
        res = msat_solve(self.env)
        assert res != MSAT_UNKNOWN
        return res == MSAT_SAT


    def solveAssuming(self, assumption):
        term = self.converter.convert(assumption)
        assert self.converter.safeState()

        if MSAT_ERROR_TERM(term):
            print("There was an error in formula assertion. The reported message is:")
            print("'" + msat_last_error_message(self.env) + "'")
            exit(1)

        res = msat_solve_with_assumptions(self.env, [term])
        assert res != MSAT_UNKNOWN
        return res == MSAT_SAT


    def solveAssuming(self, assumption):
        term = self.converter.convert(assumption)
        assert self.converter.safeState()

        if MSAT_ERROR_TERM(term):
            print("There was an error in formula assertion. The reported message is:")
            print("'" + msat_last_error_message(self.env) + "'")
            exit(1)

        res = msat_solve_with_assumptions(self.env, [term])
        assert res != MSAT_UNKNOWN
        return res == MSAT_SAT


    def solveAssuming(self, assumption):
        term = self.converter.convert(assumption)
        assert self.converter.safeState()

        if MSAT_ERROR_TERM(term):
            print("There was an error in formula assertion. The reported message is:")
            print("'" + msat_last_error_message(self.env) + "'")
            exit(1)

        res = msat_solve_with_assumptions(self.env, [term])
        assert res != MSAT_UNKNOWN
        return res == MSAT_SAT


    def allSat(self, important, callback):
        self.push()
        msat_all_sat(self.env, [self._var2Term(x) for x in important], callback)
        self.pop()

    def push(self):
        msat_push_backtrack_point(self.env)
        return


    def pop(self):
        msat_pop_backtrack_point(self.env)
        return

    def _var2Term(self, var):
        decl = msat_find_decl(self.env, var.name)
        titem = msat_make_term(self.env, decl, [])
        return titem

    def term2VarName(self, term):
        no = msat_term_is_not(self.env, term)
        var = msat_term_is_atom(self.env, term)
        assert no or var

        neg = False
        if no:
            neg = True
            term = msat_term_get_arg(term, 0)

        return msat_term_repr(term), neg


    def setPreferredVar(self, var):
        tvar = self.converter.convert(var)
        msat_add_preferred_for_branching(self.env, tvar)
        return


    def getModelValue(self, item):
        # if FNode.SYMBOL == item.nodeType:
        #     decl = msat_find_decl(self.env, item.name)
        #     titem = msat_make_term(self.env, decl, [])
        # else:
        titem = self.converter.convert(item)

        if MSAT_ERROR_TERM(titem):
            print("There was an error in model extraction phase 1. The reported message is:")
            print("'" + msat_last_error_message(self.env) + "'")
            exit(1)

        tval = msat_get_model_value(self.env, titem)
        if MSAT_ERROR_TERM(tval):
            print("There was an error in model extraction phase 2. The reported message is:")
            print("'" + msat_last_error_message(self.env) + "'")
            exit(1)

        if msat_term_is_number(self.env, tval):

            rep = msat_term_repr(tval)

            match = re.match(r"(-?\d+)/(\d+)", rep)
            if match is not None:
                return Fraction(int(match.group(1)), int(match.group(2)))
            else:
                return Fraction(int(rep), 1)

        else:
            assert msat_term_is_true(self.env, tval) or msat_term_is_false(self.env, tval)
            return msat_term_is_true(self.env, tval)

    def destroy(self):
        msat_destroy_env(self.env)
        msat_destroy_config(self.config)
        return

    pass
