from typing import *
from fractions import Fraction

class FormulaManager(object):

    def __init__(self):
        self.formulae = {}
        return

    def createNode(self, nodeType, **parameters):
        n = FNode(nodeType, parameters)
        res = self.formulae.setdefault(n,n)
        res.shared = True
        return res

    pass


# Global Formula Manager!
globalFormulaManager = FormulaManager()



class FNode(object):
    # Node types
    (FORALL, EXISTS, FUNCTION, DEFINITION, AND, OR, NOT, GE,
     PLUS, MINUS, GT, LE, LT, SYMBOL, CONSTANT, BCONSTANT, TEXTFORMULA, TIMES, DIV, EQUALS) = range(0,20)

    def __init__(self, typeId, fields, manager=globalFormulaManager):
        self.__dependencies = None
        self.__sons = None
        self.__simplification = None

        self.nodeType = typeId
        self.fields = list(fields.keys())
        self.manager = manager

        for (f,v) in fields.items():
            setattr(self, f, v)

        self.hash = self.hashFunction()
        self.shared = False
        return


    def __hash__(self):
        return self.hash

    def hashFunction(self):
        res = self.nodeType + 1
        for i in range(0, len(self.fields)):
            a = getattr(self, self.fields[i])
            res = (31 * res) ^ a.__hash__()
        if res == -1:
            res = -2
        return res


    def __eq__(self, other):
        if other is None:
            return False

        if id(self) == id(other):
            return True

        if self.shared and other.shared:
            return False

        if self.hash != other.hash:
            return False

        if self.nodeType != other.nodeType:
            return False

        #     # Two nodes of the same type must have the same fields!
        #     assert self.fields == other.fields

        for f in self.fields:
            a = getattr(self, f)
            b = getattr(other, f)

            if a != b:
                return False
        return True


    def getDependencies(self):
        if self.__dependencies is None:
            self.__dependencies = self.computeDependencies()
        return self.__dependencies


    def computeDependencies(self):
        if self.nodeType in [FNode.GE, FNode.LE, FNode.GT, FNode.LT, FNode.MINUS, FNode.PLUS, FNode.AND, FNode.OR, FNode.NOT, FNode.TIMES, FNode.DIV, FNode.EQUALS]:
            res = set()
            for s in self.getSons():
                res.update(s.getDependencies())
            return res

        elif self.nodeType in [FNode.EXISTS, FNode.FORALL]:
            return self.formula.getDependencies().difference(self.variables)

        elif self.nodeType == FNode.SYMBOL:
            return frozenset([self])

        elif self.nodeType in (FNode.CONSTANT, FNode.BCONSTANT):
            return frozenset()

        elif self.nodeType == FNode.DEFINITION:
            ss = self.formula.getDependencies()
            res = set()
            for s in ss:
                if s.nodeType == FNode.SYMBOL:
                    if s.name != self.name:
                        res.add(s)
                else:
                    res.add(s)
            return res.union(self.definition.getDependencies())

        elif self.nodeType == FNode.TEXTFORMULA:
            return frozenset(self.dependencies)

        elif self.nodeType == FNode.FUNCTION:
            res = set([self.name])
            for p in self.params:
                res.update(p.getDependencies())
            return res

        else:
            assert False
        return


    def getSons(self):
        if self.__sons is None:
            self.__sons = self.computeSons()
        return self.__sons


    def computeSons(self):
        if self.nodeType == FNode.AND:
            return self.conjunction
        if self.nodeType == FNode.OR:
            return self.disjunction
        if self.nodeType in [FNode.GE, FNode.LE, FNode.GT, FNode.LT, FNode.MINUS, FNode.PLUS, FNode.TIMES, FNode.DIV, FNode.EQUALS]:
            return [self.left, self.right]
        elif self.nodeType in [FNode.SYMBOL, FNode.CONSTANT, FNode.BCONSTANT, FNode.TEXTFORMULA]:
            return []
        elif self.nodeType in [FNode.NOT, FNode.FORALL, FNode.EXISTS, FNode.DEFINITION]:
            return [self.formula]
        elif elf.nodeType == FNode.FUNCTION:
            return self.params

        else:
            assert False
        return


    def setSimp(self, defs, simplified):
        if not defs:
            self.__simplification = simplified

        return simplified


    def simplify(self, simplifyDefinitions=False, definitions={}):
        if not simplifyDefinitions:
            s = self.__simplification
            if s is not None:
                return s

        if self.nodeType == FNode.AND:
            simp = []
            for d in self.conjunction:
                sd = d.simplify(simplifyDefinitions, definitions)
                if sd.nodeType == FNode.BCONSTANT:
                    if not sd.value:
                        return self.setSimp(simplifyDefinitions, BConstant(False, self.manager))
                elif sd.nodeType == FNode.AND:
                    simp += sd.conjunction
                else:
                    simp.append(sd)
            if len(simp) == 0:
                return self.setSimp(simplifyDefinitions, BConstant(True, self.manager))
            elif len(simp) == 1:
                return self.setSimp(simplifyDefinitions, simp[0])
            else:
                return self.setSimp(simplifyDefinitions, And(simp, self.manager))

        elif self.nodeType == FNode.OR:
            simp = []
            for d in self.disjunction:
                sd = d.simplify(simplifyDefinitions, definitions)
                if sd.nodeType == FNode.BCONSTANT:
                    if sd.value:
                        return self.setSimp(simplifyDefinitions, BConstant(True, self.manager))
                elif sd.nodeType == FNode.OR:
                    simp += sd.disjunction
                else:
                    simp.append(sd)
            if len(simp) == 0:
                return self.setSimp(simplifyDefinitions, BConstant(False, self.manager))
            elif len(simp) == 1:
                return self.setSimp(simplifyDefinitions, simp[0])
            else:
                return self.setSimp(simplifyDefinitions, Or(simp, self.manager))

        elif self.nodeType == FNode.NOT:
            sl = self.formula.simplify(simplifyDefinitions, definitions)
            if sl.nodeType == FNode.BCONSTANT:
                l = sl.value
                return self.setSimp(simplifyDefinitions, BConstant(not l, self.manager))
            elif sl.nodeType == FNode.NOT:
                return self.setSimp(simplifyDefinitions, sl.formula)
            return self.setSimp(simplifyDefinitions, Not(sl, self.manager))

        elif self.nodeType == FNode.SYMBOL:
            if simplifyDefinitions and self.name in definitions:
                return definitions[self.name]
            return self.setSimp(simplifyDefinitions, self)

        elif self.nodeType == FNode.EQUALS:
            sl = self.left.simplify(simplifyDefinitions, definitions)
            sr = self.right.simplify(simplifyDefinitions, definitions)
            if sl.nodeType == FNode.CONSTANT and sr.nodeType == FNode.CONSTANT:
                l = sl.value
                r = sr.value
                return self.setSimp(simplifyDefinitions, BConstant(l == r, self.manager))
            elif sl == sr:
                return self.setSimp(simplifyDefinitions, BConstant(True, self.manager))
            else:
                return self.setSimp(simplifyDefinitions, Equals(sl, sr, self.manager))

        elif self.nodeType == FNode.GE:
            sl = self.left.simplify(simplifyDefinitions, definitions)
            sr = self.right.simplify(simplifyDefinitions, definitions)
            if sl.nodeType == FNode.CONSTANT and sr.nodeType == FNode.CONSTANT:
                l = sl.value
                r = sr.value
                return self.setSimp(simplifyDefinitions, BConstant(l >= r, self.manager))
            return self.setSimp(simplifyDefinitions, GE(sl, sr, self.manager))

        elif self.nodeType == FNode.LE:
            sl = self.left.simplify(simplifyDefinitions, definitions)
            sr = self.right.simplify(simplifyDefinitions, definitions)
            if sl.nodeType == FNode.CONSTANT and sr.nodeType == FNode.CONSTANT:
                l = sl.value
                r = sr.value
                return self.setSimp(simplifyDefinitions, BConstant(l <= r, self.manager))
            return self.setSimp(simplifyDefinitions, LE(sl, sr, self.manager))

        elif self.nodeType == FNode.GT:
            sl = self.left.simplify(simplifyDefinitions, definitions)
            sr = self.right.simplify(simplifyDefinitions, definitions)
            if sl.nodeType == FNode.CONSTANT and sr.nodeType == FNode.CONSTANT:
                l = sl.value
                r = sr.value
                return self.setSimp(simplifyDefinitions, BConstant(l > r, self.manager))
            return self.setSimp(simplifyDefinitions, GT(sl, sr, self.manager))

        elif self.nodeType == FNode.LT:
            sl = self.left.simplify(simplifyDefinitions, definitions)
            sr = self.right.simplify(simplifyDefinitions, definitions)
            if sl.nodeType == FNode.CONSTANT and sr.nodeType == FNode.CONSTANT:
                l = sl.value
                r = sr.value
                return self.setSimp(simplifyDefinitions, BConstant(l < r, self.manager))
            return self.setSimp(simplifyDefinitions, LT(sl, sr, self.manager))

        elif self.nodeType == FNode.CONSTANT:
            return self.setSimp(simplifyDefinitions, self)

        elif self.nodeType == FNode.BCONSTANT:
            return self.setSimp(simplifyDefinitions, self)

        elif self.nodeType == FNode.FORALL:
            sf = self.formula.simplify(simplifyDefinitions, definitions)
            var = self.variables.intersection(sf.getDependencies())
            if len(var) == 0:
                return self.setSimp(simplifyDefinitions, sf)
            return self.setSimp(simplifyDefinitions, ForAll(var, sf, self.manager))

        elif self.nodeType == FNode.EXISTS:
            sf = self.formula.simplify(simplifyDefinitions, definitions)
            var = self.variables.intersection(sf.getDependencies())
            if len(var) == 0:
                return self.setSimp(simplifyDefinitions, sf)
            return self.setSimp(simplifyDefinitions, Exists(var, sf, self.manager))

        elif self.nodeType == FNode.DEFINITION:
            body = self.definition.simplify(simplifyDefinitions, definitions)
            if simplifyDefinitions:
                old = None
                if self.name in definitions:
                    old = definitions[self.name]

                definitions[self.name] = body
                res = self.formula.simplify(simplifyDefinitions, definitions)

                assert res.nodeType != FNode.DEFINITION

                if old is None:
                    del definitions[self.name]
                else:
                    definitions[self.name] = old

                # nd = dict(definitions)
                # nd[self.name] = body
                # res = self.formula.simplify(simplifyDefinitions, nd)
                return res
            else:
                return self.setSimp(simplifyDefinitions,
                                    Definition(self.name,
                                               body,
                                               self.formula.simplify(simplifyDefinitions, definitions),
                                               self.manager))


        elif self.nodeType == FNode.PLUS:
            sl = self.left.simplify(simplifyDefinitions, definitions)
            sr = self.right.simplify(simplifyDefinitions, definitions)
            if sl.nodeType == FNode.CONSTANT and sr.nodeType == FNode.CONSTANT:
                l = sl.value
                r = sr.value
                return self.setSimp(simplifyDefinitions, Constant(l + r, self.manager))

            if sl.nodeType == FNode.CONSTANT:
                l = sl.value
                if l == Fraction(0, 1):
                    return self.setSimp(simplifyDefinitions, sr)

            if sr.nodeType == FNode.CONSTANT:
                l = sr.value
                if l == Fraction(0, 1):
                    return self.setSimp(simplifyDefinitions, sl)


            return self.setSimp(simplifyDefinitions, Plus(sl, sr, self.manager))

        elif self.nodeType == FNode.TIMES:
            sl = self.left.simplify(simplifyDefinitions, definitions)
            sr = self.right.simplify(simplifyDefinitions, definitions)
            if sl.nodeType == FNode.CONSTANT and sr.nodeType == FNode.CONSTANT:
                l = sl.value
                r = sr.value
                return self.setSimp(simplifyDefinitions, Constant(l * r, self.manager))

            if sl.nodeType == FNode.CONSTANT:
                l = sl.value
                if l == Fraction(1, 1):
                    return self.setSimp(simplifyDefinitions, sr)
                elif l == Fraction(0, 1):
                    return self.setSimp(simplifyDefinitions, Constant(0, self.manager))

            if sr.nodeType == FNode.CONSTANT:
                r = sr.value
                if r == Fraction(1, 1):
                    return self.setSimp(simplifyDefinitions, sl)
                elif r == Fraction(0, 1):
                    return self.setSimp(simplifyDefinitions, Constant(0, self.manager))

            return self.setSimp(simplifyDefinitions, Times(sl, sr, self.manager))

        elif self.nodeType == FNode.DIV:
            sl = self.left.simplify(simplifyDefinitions, definitions)
            sr = self.right.simplify(simplifyDefinitions, definitions)

            if sl.nodeType == FNode.CONSTANT and sr.nodeType == FNode.CONSTANT:
                l = sl.value
                r = sr.value
                return self.setSimp(simplifyDefinitions, Constant(l / r, self.manager))

            if sl.nodeType == FNode.CONSTANT:
                l = sl.value
                if l == Fraction(0, 1):
                    return self.setSimp(simplifyDefinitions, Constant(0, self.manager))

            if sr.nodeType == FNode.CONSTANT:
                r = sr.value
                if r == Fraction(1, 1):
                    return self.setSimp(simplifyDefinitions, sl)
                elif r == Fraction(0, 1):
                    assert False, "Division by 0"

            return self.setSimp(simplifyDefinitions, Div(sl, sr, self.manager))


        elif self.nodeType == FNode.MINUS:
            sl = self.left.simplify(simplifyDefinitions, definitions)
            sr = self.right.simplify(simplifyDefinitions, definitions)
            if sl.nodeType == FNode.CONSTANT and sr.nodeType == FNode.CONSTANT:
                l = sl.value
                r = sr.value
                return self.setSimp(simplifyDefinitions, Constant(l - r, self.manager))

            if sr.nodeType == FNode.CONSTANT:
                l = sr.value
                if l == Fraction(0, 1):
                    return self.setSimp(simplifyDefinitions, sl)

            return self.setSimp(simplifyDefinitions, Minus(sl, sr, self.manager))

        elif self.nodeType == FNode.FUNCTION:
            return self.setSimp(simplifyDefinitions, self)

        elif self.nodeType == FNode.TEXTFORMULA:
            return self.setSimp(simplifyDefinitions, self)

        assert False
        return

    pass



def ForAll(variables, formula, manager=globalFormulaManager):
    n = manager.createNode(FNode.FORALL, variables=tuple(variables), formula=formula)
    return n


def Exists(variables, formula, manager=globalFormulaManager):
    n = manager.createNode(FNode.EXISTS, variables=tuple(variables), formula=formula)
    return n


def Function(vname, params, manager=globalFormulaManager):
    n = manager.createNode(FNode.FUNCTION, name=vname, params=list(params))
    return n


def Definition(name, definition, formula, manager=globalFormulaManager):
    n = manager.createNode(FNode.DEFINITION, name=name, definition=definition, formula=formula)
    return n


def And(conjunction, manager=globalFormulaManager):
    n = manager.createNode(FNode.AND, conjunction=tuple(conjunction))
    return n


def Or(disjunction, manager=globalFormulaManager):
    n = manager.createNode(FNode.OR, disjunction=tuple(disjunction))
    return n


def Not(formula, manager=globalFormulaManager):
    n = manager.createNode(FNode.NOT, formula=formula)
    return n

def Implies(left, right, manager=globalFormulaManager):
    r = Or([Not(left, manager), right], manager)
    return r

def Iff(left, right, manager=globalFormulaManager):
    r = Or([And([left, right], manager),
            And([Not(left, manager), Not(right, manager)], manager)], manager)
    return r

def GE(left, right, manager=globalFormulaManager):
    n = manager.createNode(FNode.GE, left=left, right=right)
    return n


def Plus(left, right, manager=globalFormulaManager):
    n = manager.createNode(FNode.PLUS, left=left, right=right)
    return n


def Minus(left, right, manager=globalFormulaManager):
    n = manager.createNode(FNode.MINUS, left=left, right=right)
    return n


def Times(left, right, manager=globalFormulaManager):
    n = manager.createNode(FNode.TIMES, left=left, right=right)
    return n

def Div(left, right, manager=globalFormulaManager):
    n = manager.createNode(FNode.DIV, left=left, right=right)
    return n

def Equals(left, right, manager=globalFormulaManager):
    n = manager.createNode(FNode.EQUALS, left=left, right=right)
    return n

def GT(left, right, manager=globalFormulaManager):
    n = manager.createNode(FNode.GT, left=left, right=right)
    return n


def LE(left, right, manager=globalFormulaManager):
    n = manager.createNode(FNode.LE, left=left, right=right)
    return n


def LT(left, right, manager=globalFormulaManager):
    n = manager.createNode(FNode.LT, left=left, right=right)
    return n


def Symbol(name, typename=Types.BOOL, manager=globalFormulaManager):
    n = manager.createNode(FNode.SYMBOL, name=name, typename=typename)
    return n


def Constant(value, manager=globalFormulaManager):
    if type(value) == Fraction:
        val = value
    elif type(value) == tuple:
        val = Fraction(value[0], value[1])
    elif type(value) == long or type(value) == int:
        val = Fraction(value, 1)
    elif type(value) == float:
        val = Fraction.from_float(value)
    else:
        assert False, "Invalid type in constant. The type was:" +str(type(value))

    n = manager.createNode(FNode.CONSTANT, value=val)
    return n

def BConstant(value, manager=globalFormulaManager):
    assert type(value) == bool
    n = manager.createNode(FNode.BCONSTANT, value=value)
    return n


def TextFormula(text, dependencies, manager=globalFormulaManager):
    n = manager.createNode(FNode.TEXTFORMULA, text=str(text), dependencies=frozenset(dependencies))
    return n



class SymbolTable(object):

    def __init__(self):
        self.elements = {}
        self._fresh_guess = 0
        return


    def createNewSymbol(self, name, typename):
        assert name not in self.elements
        v = Symbol(name, typename)
        self.elements[name] = v
        return v


    def createNewFreshSymbol(self, typename, base="FV%d"):
        count = self._fresh_guess
        while (base % count) in self.elements:
            count = count + 1

        name = (base % count)
        self._fresh_guess = count + 1
        v = self.createNewSymbol(name, typename)
        assert v is not None
        return v


    def getSymbolByName(self, name):
        if name in self.elements:
            return self.elements[name]
        return None


    def getOrCreateSymbol(self, name, typename):
        s = self.getSymbolByName(name)
        if s is None:
            s = self.createNewSymbol(name, typename)
        return s

    def defineSymbolGlobally(self, symbolName, definition):
        self.definitions[symbolName] = definitions
        return

    pass



class Environment(object):

    def __init__(self):
        self.symbolTable = SymbolTable()
        return

    pass
