#!/usr/bin/env python

import argparse
import os.path
import sys

sys.path.append(os.path.join(os.path.dirname(sys.argv[0]), "lib"))
sys.path.append(os.path.join(os.path.dirname(sys.argv[0]), "bin", "pylib"))
sys.setrecursionlimit(100000)

from structures import Problem, AtomicConstraint
from typing import Types
from formula import Environment, And, Or, Constant, BConstant, GE, LE, Minus
from extractors.solvers.msat import MathSAT5Solver
import common
from encoders.printers import printFormula

class PVYS(object):

    def __init__(self, environment):
        self.environment = environment
        self.fvars = None
        return


    def solve(self, problem, model=False):
        st = self.environment.symbolTable

        Cs = set()
        Cc = set()
        Ce = set()

        for c in problem.constraints:
            if c.isSimple() or not c.hasUncontrollables():
                # all simple constraints + constraints on controllable variables
                Cs.add(c)
            elif c.contingent:
                # Disjunctive contingent constraints
                Cc.add(c)
            else:
                # Disjunctive free constraints with at least one uncontrollable
                Ce.add(c)


        self.fvars = []
        for tp in problem.timePoints.values():
            if tp.controllable:
                v = st.getOrCreateSymbol(tp.name, Types.REAL)
                self.fvars.append(v)

        A = self.get_new_solver()
        res = self.dtpu_sc(A, set(), Cs, Cc, Ce)
        A.destroy()

        sat = res is not None
        if model and sat:
            print "--- Strong Schedule ---"
            for v in self.fvars:
                val = res.getModelValue(v)
                print "Schedule %s at %s" % (v.name, val)
            print "-----------------------"
            print ""

        if sat:
            res.destroy()

        self.fvars = None

        return sat

        # val = self.is_consistent(res, model=model)

        # self.solver.destroy()
        # self.solver = None

        # if model:
        #     if val[0]:
        #         print "--- Strong Schedule ---"
        #         for k, v in val[1].items():
        #             print "Schedule %s at %s" % (k.name, v)
        #         print "-----------------------"
        #         print ""
        #     return val[0]
        # else:
        #     return val

    def select_variable(self, s):
        assert len(s) > 0
        for v in s:
            return v

    def get_new_solver(self, initial_formula=None):
        A = MathSAT5Solver(self.environment)
        for v in self.fvars:
            A.declareVariable(v)

        if initial_formula is not None:
            A.addAssertion(initial_formula)

        return A

    def disjuncts(self, c):
        return c.atoms


    def sc_transform(self, prob, c):
        def get_contingency(t):
            for a in prob:
                if a.contingent and a.dest == t:
                    return (a.source, a.lowerBound, a.upperBound)
            return (None, None, None)

        def plus(a, b):
            assert b != common.INFINITY and b != common.MINUS_INFINITY
            if a == common.INFINITY or a == common.MINUS_INFINITY:
                return a

            return a + b


        toprocess = []
        if c.contingent:
            for a in prob:
                if a != c and (a.source == c.dest or a.dest == c.dest):
                    toprocess.append(a)
        else:
            toprocess.append(c)

        result = set()
        for p in toprocess:
            if p.source.controllable and p.dest.controllable:
                result.add(c)
            elif p.dest.controllable:
                # Free of type 2
                (s, l, u) = get_contingency(p.source)
                if s is not None:
                    res = AtomicConstraint(s, p.dest,
                                           plus(p.lowerBound, u),
                                           plus(p.upperBound, l),
                                           False)
                    result.add(res)

            elif p.source.controllable:
                # Free of type 3
                (s, l, u) = get_contingency(p.dest)
                if s is not None:
                    res = AtomicConstraint(p.source, s,
                                           plus(p.lowerBound, -l),
                                           plus(p.upperBound, -u),
                                           False)
                    result.add(res)

            else:
                # Free of type 4
                (s1, l1, u1) = get_contingency(p.dest)
                (s2, l2, u2) = get_contingency(p.source)
                if s1 is not None and s2 is not None:
                    res = AtomicConstraint(s1, s2,
                                           plus(p.lowerBound, u2 - l1),
                                           plus(p.upperBound, l2 - u1),
                                           False)
                    result.add(res)
        return result


    def to_formulae(self, s):
        st = self.environment.symbolTable
        res = []

        for a in s:
            vs = st.getSymbolByName(a.source.name)
            vd = st.getSymbolByName(a.dest.name)
            l = a.lowerBound
            u = a.upperBound

            if l != common.MINUS_INFINITY:
                res.append(GE(Minus(vd,vs), Constant(l)))
            if u != common.INFINITY:
                res.append(LE(Minus(vd,vs), Constant(u)))
        return res


    def dtpu_sc(self, A, Ac, Cs, Cc, Ce, formula=None):
        if formula is None:
            formula = set()

        if len(Cs) == 0:
            G = self.get_new_solver(And(formula))
            res = self.all_paths_sc(A, Ac, Cc, Ce, G, formula, formula)
            # printFormula(And(res))
            if G.solve():
                return G
            else:
                G.destroy()
        else:
            Ci = self.select_variable(Cs)

            Cs1 = Cs - set([Ci])

            for cij in self.disjuncts(Ci):
                Ac1 = Ac | set([cij])

                # Conjoin new constraints in STP A
                f = And(self.to_formulae(self.sc_transform(Ac1, cij)))

                A.push()
                A.addAssertion(f)

                if A.solve():
                    S = self.dtpu_sc(A, Ac1, Cs1, Cc, Ce, formula | set([f]))
                    if S is not None:
                        A.pop()
                        return S
                A.pop()

        return None


    def all_paths_sc(self, A, Ac, Cc, Ce, G, Aformula, Gformula):
        if len(Cc) == 0:
            f = self.satisfy_ce(A, Ac, Ce, Aformula)
            G.addAssertion(f)
            return Gformula | set([f])

        else:
            Ci = self.select_variable(Cc)
            Cc1 = Cc - set([Ci])
            for cij in self.disjuncts(Ci):
                Ac1 = Ac | set([cij])
                A.push()
                f = And(self.to_formulae(self.sc_transform(Ac1, cij)))
                A.addAssertion(f)
                Aformula1 = Aformula | set([f])
                if A.solve():
                    Gformula = self.all_paths_sc(A, Ac1, Cc1, Ce, G, Aformula1, Gformula)
                    if not G.solve():
                        A.pop()
                        return BConstant(False)
                else:
                    A.pop()
                    return BConstant(False)

                A.pop()

        return Gformula


    def satisfy_ce(self, A, Ac, Ce, Aformula):
        Hformula = set()
        if len(Ce) == 0:
            return And(Aformula)
        else:
            Ci = self.select_variable(Ce)
            Ce1 = Ce - set([Ci])
            for cij in self.disjuncts(Ci):
                f = And(self.to_formulae(self.sc_transform(Ac, cij)))
                A.push()
                A.addAssertion(f)
                if A.solve():
                    Hformula = Hformula | set([self.satisfy_ce(A, Ac, Ce1, Aformula | set([f]))])
                A.pop()
        return Or(Hformula)

#End of PVYS




# Main script
if __name__ == "__main__":

    parser = \
        argparse.ArgumentParser(description='PVYS Strong Controllability Solver')

    parser.add_argument('inputFile', metavar='inputFile', type=str,
                        help='The probmem to solve')
    parser.add_argument('--model', '-m', action='store_true',
                        help='Print a strong controllability strategy')

    args = parser.parse_args()




    env = Environment()

    tn = Problem()
    tn.fromFile(args.inputFile)


    s = PVYS(env)
    res = s.solve(tn, model=args.model)
    if res:
        print "sat"
    else:
        print "unsat"
